# Copyright (c) 2003 Sylvain Thnault (thenault@nerim.net)
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Python Abstract Syntax Tree New Generation

Define objects to represent Python code
 
"""
import sys
import token
import symbol
from types import ListType, TupleType
from os.path import dirname

from logilab.common.tree import Node
from logilab.common.modutils import load_module_from_name

__author__ = u"Sylvain Thnault"
__revision__ = "$Id: astng_objects.py,v 1.8 2003/06/10 11:25:37 syt Exp $"

#from pprint import pprint
#from logilab.pypasax.pypasax import cvrtr

# utilities ###################################################################

def print_tokens(tok_list):
    for type, tok in tok_list:
        print 'token.'+token.tok_name[type], tok
    
def debuild(tuple):
    """
    reverse tuple to string
    """
    if type(tuple[1]) in (ListType, TupleType):
        result = ''
        for child in tuple[1:]: 
            result = '%s%s' % (result, debuild(child))
        return result
    else:
        return tuple[1]

def clean(tuple):
    """
    reverse ast tuple to a list of tokens
    merge sequences (token.NAME, token.DOT, token.NAME)
    """
    result = []
    last = None
    for couple in _clean(tuple):
        if couple[0] == token.NAME and last == token.DOT:
            result[-1][1] += couple[1]
        elif couple[0] == token.DOT and last == token.NAME:
            result[-1][1] += couple[1]
        else:
            result.append(couple)
        last = couple[0]
    return result

def _clean(tuple):
    """ transform the ast into as list of tokens (i.e. final elements)
    """
    if type(tuple[1]) in (ListType, TupleType):
        v = []
        for c in tuple[1:]:
                v += _clean(c)
        return v
    else:
        return [list(tuple[:2])]

# keywords that may appears in a list of tokens
KEYWORDS = ('and', 'or', 'not', 'for', 'in', 'if', 'is', 'lambda')

def get_consumed_names(tokens, keywords=KEYWORDS):
    """ given a list of (type, token), return the list of consumed variables
    (token.NAME)
    this will typically remove named arguments in function call.
    """
    result = []
    level = 0
    skip = 0
    for i in range(len(tokens)):
        type, tok = tokens[i]
        if type == token.NAME and not tok in keywords and not skip:
            if len(tokens) <= i+1:
                result.append(tok.split('.')[0])
            elif tokens[i+1][0] != token.EQUAL:
                result.append(tok.split('.')[0])
        elif tok == 'lambda':
            skip = 1
        elif skip:
            if type in (token.LSQB, token.LPAR):
                level += 1
            elif type in (token.RSQB, token.RPAR):
                level -= 1
            elif not level and type == token.COMMA:
                skip = 0
    return result

def get_lhs_consumed_names(tokens, keywords=KEYWORDS):
    """ given a list of (type, token), return the list of consumed variables
    as if the token was the lhs part of an expression.
    this will typically remove named arguments in function call.
    """
    result = []
    keep = 0
    for type, tok in tokens:
        if type == token.NAME and not tok in keywords:
            if keep or tok.find('.') > -1:
                result.append(tok.split('.', 1)[0])
        elif type in (token.LSQB, token.LPAR):
            keep += 1
        elif type in (token.RSQB, token.RPAR):
            keep -= 1
    return result

def get_names(tokens, result=None):
    """
    get a list of names from a list of tokens (i.e comma separated values)
    """
    result = []
    new = None
    skip = 0
    for type, tok in tokens:
        if type == token.NAME and not skip:
            if new is None:
                result.append(tok)
            else:
                result[-1] += '.%s' % tok
                new = None
        elif type == symbol.trailer:
            new = 1
        elif type == token.LSQB:
            skip += 1
            new = None
        elif type == token.RSQB:
            skip -= 1
    return result


def tokens_to_string(tokens):
    return ''.join([tok for type, tok in tokens])
                   
def display_name(stmt):
    """ return the name of the class lowered and without the 'Stmt' suffix
    """
    return stmt.__class__.__name__[:-4].lower()


# MixIns ######################################################################

class DocumentedMixIn:
    """ mixin for node which support doc string """


class FrameMixIn:
    """ FrameMixin provide the following methods :
    
    _ locals     : return the dictionary of local variables
    _ add_local  : add a name to the local variables
    _ globals     : return the dictionary of global variables
    """
    
    # list of attributes which may be acquired to the bound object
    ACQUIRED = ()
    
    def __init__(self, object=None):
        self._locals = {}
        self.object = object
        
    def add_local(self, name, stmt):
        """ add a name to the local dictionary

        optional stmt is the statement where the local has been introduced
        """
        if not self._locals.has_key(name):
            self._locals[name] = stmt
            
    def __getattr__(self, attr):
        if self.object is None:
            raise AttributeError(attr)
        if not attr in self.ACQUIRED:
            raise AttributeError(attr)            
        try:
            return getattr(self.object, attr)
        except AttributeError:
            raise AttributeError(attr)

    def locals(self):
        """ return the locals dictionary """
        return self._locals
    
    def globals(self):
        """ return the globals dictionary (i.e. the parent module locals) """
        return self.parent.globals()

    def docstring(self):
        """ return the docstring if any, else None
        """
        if self.object is not None:
            return self.object.__doc__
        if self.children and isinstance(self.children[0], StringStmt):
            return self.children[0].value

    def set_object(self):
        """ try to get self.object """
        if self.object is not None:
            return
        parent_obj = getattr(self.parent, 'object', None)
        if parent_obj is not None:
            self.object = getattr(parent_obj, self.__name__, None)

    def get_frame(self):
        """ return the first object inheriting from FrameMixIn form here to the
        root of the tree
        """
        return self

    
class SubFrameMixIn:
    """ SubFrameMixIn provides the same methods as FrameMixIn but delegate
    works to its parent node
    """
    def add_local(self, name, stmt):
        """ add a name to the local dictionary

        optional stmt is the statement where the local has been introduced
        """
        self.parent.add_local(name, stmt)

    def locals(self):
        """ return the locals dictionary """
        return self.parent.locals()

    def globals(self):
        """ return the globals dictionary (i.e. the parent module locals) """
        return self.parent.globals()
    
    def get_frame(self):
        """ return the first object inheriting from FrameMixIn form here to the
        root of the tree
        """
        return self.parent.get_frame()
    



# abstract classes ############################################################

class Statement(Node):
    def __init__(self, num, line):
        Node.__init__(self)
        # statement line
        self.line = line
        # statement number
        self.num = num
        
    def __str__(self):
        return  '%s (line %s)' % (str(self.__class__), self.line)
        
    def __repr__(self, indent=''):
        s = '%s%s\n' % (indent, str(self))
        for c in self.children:
            s += c.__repr__(indent+'    ')
        return s

    def accept(self, visitor, *args, **kwargs):
        func = getattr(visitor, 'visit_%s' % self.__class__.__name__.lower())
        return func(self, *args, **kwargs)

    def append(self, child):
        Node.append(self, child)
        if hasattr(child, '__name__'):
            self.add_local(child.__name__, child)
        
    def last_statement(self):
        """ return the last children statement (i.e. the right most leaf of
        the statements tree)
        """
        if self.children:
            return self.children[-1].last_statement()
        return self

class FinalStatement(Statement):
    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        pass
    
    def __str__(self):
        return display_name(self)


class ConditionalStatement(Statement):
    def __init__(self, num, line):
        Statement.__init__(self, num, line)
        self.tokens = ()

    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        if len(tuple) > 2:
            self.tokens = clean(tuple[2])
            handle_list_comprehension(self, self.tokens)
            
    def __str__(self):
        s = display_name(self)
        return '%s %s:' % (s, tokens_to_string(self.tokens))


class KWStatement(Statement):
    def __init__(self, num, line):
        Statement.__init__(self, num, line)
        self.keyword = display_name(self)
        self.tokens = []

    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        if len(tuple) > 2:
            self.tokens = clean(tuple[2])
            handle_list_comprehension(self, self.tokens)

    def __str__(self):
        return '%s %s' % (self.keyword, tokens_to_string(self.tokens))

class KWTupleStatement(KWStatement):
    """
    idem as KW statement but may have multiple rhs values separated by comma
    """
    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        for tuple in tuple[2:]:
            self.tokens += clean(tuple)
        if self.tokens:
            handle_list_comprehension(self, self.tokens)



# concrete classes ############################################################

    
class Module(Statement, FrameMixIn): 
    """ a module represent a single file of python code

    this is the only object which is surely bound to a real object
    """

    ACQUIRED = ('__name__',)
    
    def __init__(self, object):
        Statement.__init__(self, 0, 0)
        FrameMixIn.__init__(self, object)
        self.file = getattr(object, '__file__', None)
        self.pure_python = 1
        
    def globals(self):
        """ overriden form FrameMixIn to return the globals dictionary
        """
        return self.locals()

    def __str__(self):
        return '<Module %s (%s) at %s>' % (self.__name__, self.file, id(self))
    
    def __repr__(self):
        s = ''
        for c in self.children:
            s += c.__repr__()
        return s
    
    def is_pure_python(self):
        """ return true if module is a pure python module """
        return self.pure_python
    
    def is_package(self):
        """ return true if module is actually a package
        """
        return hasattr(self.object, '__path__')


class ClassDef(Statement, FrameMixIn):
    """ a class definition statement

    attributes :
      _ __name__       : the name of this class
      _ __bases__      : list of base classes in order
      _ bases          : list of base classes name in order (as it was defined
                         in the source code)
      _ instance_attrs : dictionnary of detected instance attributes
    """
    ACQUIRED = ('__bases__',)
    
    def __init__(self, num, line, name=None, object=None):
        assert name or object
        Statement.__init__(self, num, line)
        FrameMixIn.__init__(self, object)
        self.__name__ = name or object.__name__
        self.bases = []
        self.instance_attrs = {}
        if object is not None:
            for classe in object.__bases__:
                self.bases.append(classe.__name__)

    def add_base_class(self, name):
        """ add a string representing a base class for this class """
        self.bases.append(name)

    def decode(self, tuple):
        """ fill instance according to an ast tuple if necessary
        (no living class)
        """
        assert self.parent is not None
        self.set_object()
        if tuple[3][0] == token.LPAR:
            for base in get_names(clean(tuple[4])):
                self.add_base_class(base)
        if self.object is None:
            # __bases__ should be linked to real base class
            self.__bases__ = []
            for base in self.bases:
                try:
                    stmt = self.globals()[base]
                except:
                    continue
                if isinstance(stmt, ImportStmt):
                    i = stmt.asnames.index(base)
                    base = stmt.names[i]
                    if stmt.base:
                        base = '%s.%s' % (stmt.base, base)
                    
                    sys.path.insert(0, dirname(self.root().file))
                    try:
                        self.__bases__.append(load_module_from_name(base, globals()))
                    except:
                        print >>sys.stderr, 'Unable to load base class %s of %s' % (
                            base, self.__name__)
                    del sys.path[0]
                elif isinstance(stmt, ClassDef):
                    try:
                        self.__bases__.append(getattr(self.root().object, base))
                    except:
                        continue
                 
    def __str__(self):
        if self.__bases__:
            return  'class %s(%s):' % (self.__name__, ', '.join(self.bases))
        else:
            return  'class %s:' % self.__name__

    def __repr__(self, indent=''):
        s = Statement.__repr__(self, indent)
        return s + '\n'


class FuncDef(Statement, FrameMixIn): 
    """ a function definition statement

    attributes :
      _ name     : the name of this function
      _ args     : list of arguments name in order
      _ defaults : dictionnary for default arguments'value if any
      _ flags    : dictionnary for arguments'flag if any (i.e. * or ** arguments)
    """

    def __init__(self, num, line, name=None, object=None):
        assert name or object
        Statement.__init__(self, num, line)
        FrameMixIn.__init__(self, object)
        self.__name__ = name or object.__name__
        self.args = []
        self.raw_defaults = []
        if object is not None:
            self.init_from_object()

    ACQUIRED = ('co_flags', 'func_defaults')

    def __getattr__(self, attr):
        if self.object is None:
            if attr == 'co_nlocals':
                return len(self.locals())
            raise AttributeError(attr)
        try:
            return getattr(self.object, attr)
        except AttributeError:
            try:
                return getattr(self.object.func_code, attr)
            except AttributeError:
                raise AttributeError(attr)
            
    def is_method(self):
        return isinstance(self.parent, ClassDef)
    
    def add_argument(self, name):
        """ add an argument for this function """
        self.args.append(name)
        self.add_local(name, self)

    def set_object(self):
        FrameMixIn.set_object(self)
        # be sure object is a function if given
        if hasattr(self.object, 'im_func'):
            self.object = self.object.im_func
        
    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        assert self.parent is not None, 'object must be appended to its parent \
before being decoded'
        self.set_object()
        if self.object is None:
            self.co_flags = 0
            self.func_defaults = []
            if tuple[3][2][0] == symbol.varargslist:
                flag = None
                for tuple in tuple[3][2][1:]:
                    if tuple[0] == symbol.fpdef:
                        self.add_argument(tuple[1][1])
                    elif tuple[0] == token.STAR:
                        self.co_flags += 4
                    elif tuple[0] == token.DOUBLESTAR:
                        self.co_flags += 8
                    elif tuple[0] == token.NAME:
                        self.add_argument(tuple[1])                    
                    elif tuple[0] != token.COMMA:
                        default = debuild(tuple)
                        try:
                            default = eval(default)
                        except:
                            pass
                        self.func_defaults.append(default)
                        self.raw_defaults.append(clean(tuple))
        else:
            self.init_from_object()
            if tuple[3][2][0] == symbol.varargslist:
                for tuple in tuple[3][2][1:]:
                    if tuple[0] in (symbol.fpdef, token.STAR,
                                    token.DOUBLESTAR, token.NAME):
                        pass
                    elif tuple[0] != token.COMMA:
                        self.raw_defaults.append(clean(tuple))
            
    def init_from_object(self):
        object = self.object
        total_args = object.func_code.co_argcount
        i = -1
        for i in range(total_args):
            self.add_argument(object.func_code.co_varnames[i])
        doublestar = (self.co_flags & 8) and (total_args - 1) or -1
        if (self.co_flags & 4) and 1 or 0:
            # there is an * argument
            i += 1
            self.add_argument(object.func_code.co_varnames[i])
        if (self.co_flags & 8) and 1 or 0:
            # there is an * argument
            i += 1
            self.add_argument(object.func_code.co_varnames[i])
            
        if object.func_defaults is None:
            object.func_defaults = ()
            
    def format_args(self):
        """ return a string which represents this function arguments """
        result = []
        total_args = len(self.args)
        doublestar = (self.co_flags & 8) and (total_args - 1) or -1
        if doublestar != -1:
            total_args -= 1
        star = (self.co_flags & 4) and (doublestar != -1 and (doublestar - 1) 
                                        or (total_args - 1)) or -1
        if star != -1:
            total_args -= 1
        for i in range(len(self.args)):
            arg = self.args[i]
            if i == doublestar:
                result.append('**%s'%arg)
            elif i == star:
                result.append('*%s'%arg)
            elif self.func_defaults and i > total_args - len(self.func_defaults):
                default = self.func_defaults[total_args - i]
                result.append('%s=%s' % (arg, default))
            else:
                result.append(arg)
        return ', '.join(result)
    
    def __str__(self):
        return 'def %s(%s):' % (self.__name__, self.format_args())

    def __repr__(self, indent=''):
        s = Statement.__repr__(self, indent)
        return s + '\n'

    
class WhileStmt(ConditionalStatement, SubFrameMixIn): pass
class IfStmt(ConditionalStatement, SubFrameMixIn): pass

class ElifStmt(ConditionalStatement, SubFrameMixIn):
    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        self.tokens = clean(tuple)
        handle_list_comprehension(self, self.tokens)
           

class ExceptStmt(ConditionalStatement, SubFrameMixIn):
    def __init__(self, num, line):
        ConditionalStatement.__init__(self, num, line)
        self.instance_tokens = ()
        
    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        ConditionalStatement.decode(self, tuple)
        if len(tuple) > 4:
            self.instance_tokens = clean(tuple[4])
            for name in get_names(self.instance_tokens):
                self.add_local(name, self)

    def __str__(self):
        s = 'except'
        if self.tokens:
            s = '%s %s' % (s, tokens_to_string(self.tokens))
        if self.instance_tokens:
            s = '%s, %s' % (s, tokens_to_string(self.instance_tokens))
        return s + ':'
    
class ForStmt(Statement, SubFrameMixIn):
    def __init__(self, num, line):
        Statement.__init__(self, num, line)
        self.variables_tokens = ()
        self.in_tokens = ()
        self.variables = []

    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        self.variables_tokens = clean(tuple[2])
        for name in get_names(self.variables_tokens):
            self.variables.append(name)
            self.parent.add_local(name.split('.')[0], self)
        self.in_tokens = clean(tuple[4])
        handle_list_comprehension(self, self.in_tokens)
    
    def __str__(self):
        for_value = tokens_to_string(self.variables_tokens)
        list_value = tokens_to_string(self.in_tokens)
        return 'for %s in %s:' % (for_value, list_value)


class TryStmt(FinalStatement, SubFrameMixIn):
    def __str__(self):
        return 'try:'
    
class ElseStmt(FinalStatement, SubFrameMixIn):
    def __str__(self):
        return 'else:'
    
class FinallyStmt(FinalStatement, SubFrameMixIn): 
    def __str__(self):
        return 'finally:'


class PassStmt(FinalStatement): pass
class ContinueStmt(FinalStatement): pass
class BreakStmt(FinalStatement): pass

class RaiseStmt(KWTupleStatement, SubFrameMixIn): pass
class PrintStmt(KWTupleStatement, SubFrameMixIn): pass
class ReturnStmt(KWTupleStatement, SubFrameMixIn): pass
class YieldStmt(KWTupleStatement, SubFrameMixIn): pass
class GlobalStmt(KWTupleStatement, SubFrameMixIn): pass

class AssertStmt(KWStatement, SubFrameMixIn):
    def __init__(self, num, line):
        KWStatement.__init__(self, num, line)
        self.msg_tokens = None

    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        KWStatement.decode(self, tuple)
        if len(tuple) > 4:
            self.msg_tokens = clean(tuple[4])
            handle_list_comprehension(self, self.msg_tokens)

    def __str__(self):
        s = KWStatement.__str__(self)
        if self.msg_tokens is not None:
            return '%s, %s' % (s, tokens_to_string(self.msg_tokens))
        return s
    

class ExecStmt(KWStatement, SubFrameMixIn): 
    def __init__(self, num, line):
        KWStatement.__init__(self, num, line)
        self.in_tokens = None

    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        KWStatement.decode(self, tuple)
        if len(tuple) > 4:
            self.in_tokens = clean(tuple[4])
            handle_list_comprehension(self, self.in_tokens)

    def __str__(self):
        s = KWStatement.__str__(self)
        if self.in_tokens is not None:
            return '%s in %s' % (s, tokens_to_string(self.in_tokens))
        return s


class ImportStmt(Statement, SubFrameMixIn):
    def __init__(self, num, line):
        Statement.__init__(self, num, line)
        self.names = []
        self.asnames = []
        self.base = ''
        self.wildcard = 0
        
    def add_imported(self, name, as_name):
        """ add an imported module
        
        name is the name of the imported object
        as_name is the alias name of the imported object (it maybe the same
        as name)
        """
        self.names.append(name)
        self.asnames.append(as_name)
        self.add_local(as_name.split('.')[0], self)

    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        if tuple[1][1] == 'from':
            self.base = prefix = debuild(tuple[2])
            i = 4
            while i < len(tuple) and (tuple[i][0] == symbol.import_as_name or
                                      tuple[i][0] == token.STAR):
                if  tuple[i][0] == token.STAR:
                    self.wildcard = 1
                    dict = {}
                    exec 'from %s import *' % prefix in dict
                    for name in dict.keys():
                        self.add_imported(name, name)
                        
                else:
                    name = debuild(tuple[i][1])
                    asname = debuild(tuple[i][-1])
                    self.add_imported(name, asname)
                i += 2
        else:
            i = 2
            while i < len(tuple) and tuple[i][0] == symbol.dotted_as_name:    
                name = debuild(tuple[i][1])
                asname = debuild(tuple[i][-1])
                self.add_imported(name, asname)
                i += 2

    def __str__(self):
        l = []
        for i in range(len(self.names)):
            name = self.names[i]
            asname = self.asnames[i]
            if name != asname:
                l.append('%s as %s' % (name, asname))
            else:
                l.append(name)
        if self.base:
            return 'from %s import %s' % (self.base, ', '.join(l))
        else:
            return 'import %s' % ', '.join(l)

        
class AssignmentStmt(Statement, SubFrameMixIn):
    """ an assignment statement (re)defining some variables """
    
    def __init__(self, num, line):
        Statement.__init__(self, num, line)
        self.variables_tokens = ()
        self.rhs_tokens = ()
        self.variables = []

    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        self.variables_tokens = clean(tuple[1])
        i = 4
        while len(tuple) > i and tuple[i][0] == token.EQUAL:
            self.variables_tokens += clean(tuple[i-2])
            self.variables_tokens += clean(tuple[i-1])
            i += 2
        # is it an assignement in a method ?
        attr_dict = None
        frame = self.get_frame()
        if isinstance(frame, FuncDef) and frame.is_method() and frame.args:
            classe = frame.parent
            attr_dict = classe.instance_attrs
            func_self = frame.args[0]
        for name in get_names(self.variables_tokens):
            self.variables.append(name)
            parts = name.split('.')
            if len(parts) == 1:
                self.add_local(parts[0], self)
                continue
            if attr_dict is not None and parts[0] == func_self:
                attr = parts[1]
                if not attr_dict.has_key(attr) or frame.__name__ == '__init__':
                    attr_dict[attr] = self
            
        self.rhs_tokens = clean(tuple[i-1])
        handle_list_comprehension(self, self.rhs_tokens)
            
    def __str__(self):
        rhs = tokens_to_string(self.rhs_tokens)
        return '%s = %s' % (tokens_to_string(self.variables_tokens), rhs)
    

class StringStmt(Statement):
    """ pure string statements like doc strings """
    
    def __init__(self, num, line):
        Statement.__init__(self, num, line)
        self.value = ''

    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        self.value = debuild(tuple)
        
    def __str__(self):
        return self.value


class ExpressionStmt(Statement, SubFrameMixIn):
    """ all others statements """
    
    def __init__(self, num, line):
        Statement.__init__(self, num, line)
        self.tokens = ()
        
    def decode(self, tuple):
        """ fill instance according to an ast tuple """
        self.tokens = clean(tuple)
        handle_list_comprehension(self, self.tokens)
        
    def __str__(self):
        return tokens_to_string(self.tokens)


def handle_list_comprehension(stmt, tokens):
    """ search for a list comprehension in tokens
    if found, add variables to locals
    """
    i = 0
    while i < len(tokens):
        type, tok = tokens[i]
        i+= 1
        if type == token.NAME and tok == 'for':
            while i < len(tokens):
                type, tok = tokens[i]
                i += 1
                if type == token.NAME:
                    if tok == 'in':
                        i += 2
                        break
                    else:
                        stmt.add_local(tok.split('.')[0], stmt)
            else:
                raise 'Internal Error' 
