# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2002-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

"""

__revision__ = "$Id: configuration.py,v 1.2 2003/06/03 09:12:51 syt Exp $"

import sys
import re
from os.path import exists
from copy import copy
from ConfigParser import ConfigParser, NoOptionError
from logilab.common.optik_ext import OptionParser, OptionGroup, Values, \
     OptionValueError, check_yn, check_csv

def choice_validator(opt_dict, name, value):
    """ validate and return a converted value for option of type 'choice'
    """
    if not value in opt_dict['choices']:
        msg = "option %s: invalid value: %r, should be in %s"
        raise OptionValueError(msg % (name, value, opt_dict['choices']))
    return value

def csv_validator(opt_dict, name, value):
    """ validate and return a converted value for option of type 'csv'
    """
    return check_csv(None, name, value)

def yn_validator(opt_dict, name, value):
    """ validate and return a converted value for option of type 'yn'
    """
    return check_yn(None, name, value)
   
VALIDATORS = {'string' : str,
              'int' : int,
              'float': float,
              'regexp': re.compile,
              'csv': csv_validator,
              'yn': yn_validator,
              'choice': choice_validator
              }

def convert(value, opt_dict, name = ''):
    """ return a validated value for an option according to its type
    
    optional argument name is only used for error message formatting
    """
    type = opt_dict['type']
    if not VALIDATORS.has_key(type):
        raise Exception('Unsupported type "%s"' % type)
    try:
        return VALIDATORS[type](opt_dict, name, value)
    except TypeError:
        try:
            return VALIDATORS[type](value)
        except:
            raise OptionValueError('%s value (%s) should be of type %s' %
                                   (name, value, type))


class OptionsManagerMixIn:
    """ MixIn to handle a configuration from both a configuration file and
    command line options
    """
    
    def __init__(self, usage, config_file=None, version=None):
        self._config_file = config_file
        # configuration file parser
        self._config_parser = ConfigParser()
        # command line parser
        self._optik_parser = OptionParser(usage=usage, version=version)
        # list of registered options providers
        self.options_providers = []
        # dictionary assocating option name to checker
        self._all_options = {}
        self._nocallback_options = {}
        
    def register_options_provider(self, provider):
        """ register an options provider """
        assert provider.priority <= 0, 'provider priority can\'t be >= 0'
        for i in range(len(self.options_providers)):
            if provider.priority > self.options_providers[i].priority:
                self.options_providers.insert(i, provider)
                break
        else:
            self.options_providers.append(provider)
        # add section to the config file
        section = provider.name
        self._config_parser.add_section(section.upper())
        # add option group to the command line parser
        group = OptionGroup(self._optik_parser,
                            title=section.capitalize(),
                            description=provider.__doc__)
        self._optik_parser.add_option_group(group)
        # add provider's specific options
        for opt_name, opt_dict in provider.options:
            opt_dict = copy(opt_dict)
            if opt_dict.has_key('action'):
                self._nocallback_options[provider] = opt_name
            else:
                opt_dict['action'] = 'callback'
                opt_dict['callback'] = self._set_provider_option
            if opt_dict.has_key('default'):
                del opt_dict['default']
            group.add_option('--' + opt_name, **opt_dict)
            self._all_options[opt_name] = provider


    def _set_provider_option(self, option, opt_name, value, parser):
        """ optik callback for option setting """
        # remove --
        opt_name = opt_name[2:]
        # trick since we can't set action='store_true' on options
        if value is None:
            value = 1
        # set option on the checker
        self._all_options[opt_name].set_option(opt_name, value)

    def generate_config(self, stream=sys.stdout):
        """ write a configuration file according to the current configuration
        into the given stream or stdout
        """
        for provider in self.options_providers:
            print >>stream, '[%s]' % provider.name.upper()
            for opt_name, opt_dict in provider.options:
                if opt_dict.get('type', None) is None:
                    continue
                attr = provider.option_name(opt_name)
                value = getattr(provider.config, attr)
                if type(value) in (type(()), type([])):
                    value = ','.join(value)
                elif hasattr(value, 'match'):
                    # compiled regex, skip value
                    value = ''
                elif opt_dict.get('type') == 'yn':
                    value = value and 'yes' or 'no'
                else:
                    value = repr(value)
                print >>stream, '%s=%s' % (opt_name, value)
            print >>stream


    # initialization methods ##################################################

    def load_file_configuration(self, config_file=None):
        """ load the configuration from file
        """
        if config_file is None:
            config_file = self._config_file
        parser = self._config_parser
        if config_file and exists(config_file):
            parser.read([config_file])
        else:
            msg = 'No config file found, using default configuration'
            print >> sys.stderr, msg
        for provider in self.options_providers:
            section = provider.name.upper()
            for opt_name, opt_dict in provider.options:
                try:
                    value = parser.get(section, opt_name)
                    # type casting
                    value = convert(value, opt_dict, opt_name)
                    provider.set_option(opt_name, value)
                except NoOptionError:
                    continue

    def load_configuration(self, **kwargs):
        """ override configuration according to given parameters
        """
        for opt_name, opt_value in kwargs.items():
            opt_name = opt_name.replace('_', '-')
            provider = self._all_options[opt_name]
            provider.set_option(opt_name, opt_value)
            
            
    def load_command_line_configuration(self, args = None):
        """ override configuration according to command line parameters

        return additional arguments
        """
        if args is None:
            args = sys.argv[1:]
        else:
            args = list(args)
        (options, args) = self._optik_parser.parse_args(args=args)
        for provider in self._nocallback_options.keys():
            config = provider.config
            for attr in config.__dict__.keys():
                value = getattr(options, attr, None)
                if value is None:
                    continue
                setattr(config, attr, value)
        return args


    # help methods ############################################################

    def add_help_section(self, title, description):
        """ add a dummy option section for help purpose """
        group = OptionGroup(self._optik_parser,
                            title=title.capitalize(),
                            description=description)
        self._optik_parser.add_option_group(group)

        
    def help(self):
        """ return the usage string for available options """
        return self._optik_parser.print_help()
    

        
class OptionsProviderMixIn:
    
    # those attributes should be overriden
    priority = -1
    name = 'default'
    options = ()

    def __init__(self):
        self.config = Values()
        for opt_name, opt_dict in self.options:
            self.set_option(opt_name, opt_dict.get('default'))

    def option_name(self, opt_name, opt_dict=None):
        """ get the config attribute corresponding to opt_name
        """
        if opt_dict is None:
            for o in self.options:
                if o[0] == opt_name:
                    opt_dict = o[1]
                    break
            else:
                raise OptionValueError('No such option %s in section %s' % (
                    opt_name, self.name))
        return opt_dict.get('dest', opt_name.replace('-', '_'))
        
    def set_option(self, opt_name, value):
        """ method called to set an option (registered in the options list)
        """
        setattr(self.config, self.option_name(opt_name), value)
    
